function demo_eigenspace (n, p)
%-------------------------------------------------------------------------
% A demo of solving
%   min F(X), s.t., X'*X = I, where X is an n-by-p matrix
%
%  This demo solves the eigenvalue problem by letting
%  F(X) = -0.5*Tr(X'*A*X); where A is a n-by-n symmetric and
%  positive definite matrix.
%-------------------------------------------------------------------------
% 
% Example1:
% demo_eigenspace ()
% 
% Example2:
% demo_eigenspace (100)
% 
% Example3:
% demo_eigenspace (1000,4)
%
% Reference:
% ----------
% Harry Oviedo Hugo Lara and Oscar Dalmau. "A Non-monotone Linear Search 
%      Method with Mixed Direction on Stiefel Manifold"
%      https://arxiv.org/abs/1702.04303
%
% Authors:
% ----------
% Harry Oviedo <harry.oviedo@cimat.mx>
% Hugo Lara    <hugo.lara.urdaneta@ufsc.br>
% Oscar Dalmau <dalmau@cimat.mx>
%
% Date: 15-Feb-2017
%-------------------------------------------------------------------------

clc
if nargin < 2  
    p = 10;
end
if nargin < 1
    n = 1000;   
end

%-------------------------------------------------------------------------
% Generating the problem 
%-------------------------------------------------------------------------

A = randn(n,n);     A = A'+A;   mineig = min(eig(A));
if mineig <= 0
    mu = abs(mineig) + 1;
else
    mu = 1;
end
A = A + mu*eye(n);  % is symmetric and positive definite

% Generating the starting point:
X0 = rand(n,p);     [X0,~] = qr(X0,0);


%-------------------------------------------------------------------------
% Solving with MATLAB eigs
% opteig.issym = 1;
% [~, D] = eigs(@funAX, n, p, 'la',opteig);
%-------------------------------------------------------------------------
[~, D] = eigs(A,p);
D = diag(D); feig = sum(D(1:p));

%-------------------------------------------------------------------------
% Solving the problem with our solver 
%-------------------------------------------------------------------------
opts.record = 0;
opts.mxitr  = 1000;
opts.xtol = 1e-7;
opts.gtol = 1e-5;
opts.ftol = 1e-13;
opts.tau  = 1e-3;

%--- Our solver ---
[X, out, F_eval, Grad]= gradient_retrac(X0, @fun, opts, A);
out.sum_max_eigenvalues = -2*out.fval; % convert the function value to the sum of eigenvalues
err = (feig - out.sum_max_eigenvalues)/(abs(feig) + 1);

%-------------------------------------------------------------------------
% Printing the results
%-------------------------------------------------------------------------
fprintf('Results\n') 
fprintf('---------------------------------------------------\n')
fprintf('   Obj. function = %7.6e\n',  out.fval);
fprintf('   Gradient norm = %7.6e \n', out.nrmG);
fprintf('   ||X^T*X-I||_F = %3.2e\n',out.feasi )
fprintf('   Iteration number = %d\n',  out.itr); 
fprintf('   Cpu time (secs) = %3.4f  \n',  out.time);
fprintf('   Number of evaluation(Obj. func) = %d\n',  out.nfe); 
fprintf('   Relative difference between obj vals = %3.2e\n',err);
%-------------------------------------------------------------------------

figure, plot(Grad)
xlabel('Iterations (k) ')
ylabel('Gradient norm (X_k)')
figure, plot(F_eval)
xlabel('Iterations (k) ')
ylabel('F(X_k)')

% Auxiliar function
  %  function AX = funAX(X)
  %      AX = A*X;
  %  end

% Objective Function (F) and Gradient (G) at X  
    function [F, G] = fun(X, A)
        AX = A*X;
        F = -0.5*trace(X'*AX);
        G = -AX;
    end
end


