function [X, out, F_eval, Grad]= gradient_retrac(X, fun, opts, varargin)
%-------------------------------------------------------------------------
% A Non-monotone Linear Search Method with Mixed Direction on Stiefel Manifold
%
%   min F(X), S.t., X'*X = I_p, where X \in R^{n,p}
%
%-------------------------------------------------------------------------
% Input:
%           X --- n by p matrix such that X'*X = I
%         fun --- objective function and its gradient:
%                 [F, G] = fun(X,  data1, data2)
%                 F, G are the objective function value and gradient, repectively
%                 data1, data2 are addtional data, and can be more
%                 Calling syntax:
%                   [X, out]= OptStiefelGBB(X0, @fun, opts, data1, data2);
%
%        opts --- option structure with fields:
%                 record = 0, no print out
%                 mxitr       max number of iterations
%                 xtol        stop control for ||X_k - X_{k-1}||
%                 gtol        stop control for the projected gradient
%                 ftol        stop control for |F_k - F_{k-1}|/(1+|F_{k-1}|)
%                             usually, max{xtol, gtol} > ftol
%                 alpha       the parameter "alpha">0 of the direction search
%                 beta        the parameter "beta">=0 of the direction search
% Output:
%           X --- solution
%         out --- output information
%      F_eval --- evaluation of the objective function in each iteration
%        Grad --- norm of gradient of lagrangean function espect to primal 
%                 variables in each iteration
%-------------------------------------------------------------------------
%
% Example: F(X) = -0.5*Tr(X'*A*X);
%
% Demo function

% function [F, G] = fun(X, A)
%   G = -(A*X);
%   F = 0.5*sum(dot(G,X,1));
% end
% 
% n = 1000; k = 6;
% A = randn(n); A = A'+A;
% opts.record = 0; 
% opts.mxitr  = 1000;
% opts.xtol = 1e-7;
% opts.gtol = 1e-5;
% opts.ftol = 1e-13;
% 
% X0 = randn(n,k);  [U,~,V] = svd(X0);  X0 = U(:,1:p)*V';
%
% [X, out, F_eval, Grad]= gradient_retrac(X0, @fun, opts, A);
% out.sum_max_eigenvalues = -2*out.fval; % convert the function value to the sum of eigenvalues
% 
% end
%-------------------------------------------------------------------------
%
% Reference: 
% H. Oviedo, H. Lara and O. Dalmau
% A Non-monotone Linear Search Method with Mixed Direction on Stiefel Manifold
% https://arxiv.org/abs/1702.04303
%
% Authors:
% Harry F. Oviedo, Hugo J. Lara and Oscar S. Dalmau
%     Harry Oviedo <harry.oviedo@cimat.mx>
%     Hugo Lara    <hugo.lara.urdaneta@ufsc.br>
%     Oscar Dalmau <dalmau@cimat.mx>
% Version 1.0 .... 2017/01
%-------------------------------------------------------------------------


%% Size information
if isempty(X)
    error('input X is an empty matrix');
else
    [n, k] = size(X);
end

if n==k
    OrthG = true;
else
    OrthG = false;
end

if isfield(opts, 'xtol')
    if opts.xtol < 0 || opts.xtol > 1
        opts.xtol = 1e-6;
    end
else
    opts.xtol = 1e-6;
end

if ~isfield(opts, 'alpha')
    opts.alpha = 1;
end

if ~isfield(opts, 'beta')
    opts.beta = 0;
end

if isfield(opts, 'gtol')
    if opts.gtol < 0 || opts.gtol > 1
        opts.gtol = 1e-4;
    end
else
    opts.gtol = 1e-4;
end

if isfield(opts, 'ftol')
    if opts.ftol < 0 || opts.ftol > 1
        opts.ftol = 1e-12;
    end
else
    opts.ftol = 1e-12;
end

% parameters for control the linear approximation in line search
if isfield(opts, 'rho')
   if opts.rho < 0 || opts.rho > 1
        opts.rho = 1e-4;
   end
else
    opts.rho = 1e-4;
end

% factor for decreasing the step size in the backtracking line search
if isfield(opts, 'eta')
   if opts.eta < 0 || opts.eta > 1
        opts.eta = 0.3;
   end
else
    opts.eta = 0.3;
end

% parameters for updating C by HongChao, Zhang
if isfield(opts, 'gamma')
   if opts.gamma < 0 || opts.gamma > 1
        opts.gamma = 0.85;
   end
else
    opts.gamma = 0.85;
end

if isfield(opts, 'tau')
   if opts.tau < 0 || opts.tau > 1e3
        opts.tau = 1e-3;
   end
else
    opts.tau = 1e-3;
end

% parameters for the  nonmontone line search by Raydan
if ~isfield(opts, 'STPEPS')
    opts.STPEPS = 1e-10;
end

if isfield(opts, 'nt')
    if opts.nt < 0 || opts.nt > 100
        opts.nt = 5;
    end
else
    opts.nt = 5;
end

if isfield(opts, 'projG')
    switch opts.projG
        case {1,2}; otherwise; opts.projG = 1;
    end
else
    opts.projG = 1;
end

if isfield(opts, 'iscomplex')
    switch opts.iscomplex
        case {0, 1}; otherwise; opts.iscomplex = 0;
    end
else
    opts.iscomplex = 0;
end

if isfield(opts, 'mxitr')
    if opts.mxitr < 0 || opts.mxitr > 2^20
        opts.mxitr = 1000;
    end
else
    opts.mxitr = 1000;
end

if ~isfield(opts, 'record')
    opts.record = 0;
end


%-------------------------------------------------------------------------------
% copy parameters
alpha = opts.alpha;
beta = opts.beta;
xtol = opts.xtol;
gtol = opts.gtol;
ftol = opts.ftol;
rho  = opts.rho;
STPEPS = opts.STPEPS;
eta   = opts.eta;
gamma = opts.gamma;
iscomplex = opts.iscomplex;
record = opts.record;

nt = opts.nt;   crit = ones(nt, 3);


%% Initial function value and gradient
% prepare for iterations
[F, G] = feval(fun, X , varargin{:});  out.nfe = 1;  
GX = G'*X;       
dtX  = G - X*GX;     nrmG  = norm(dtX, 'fro');
if ~OrthG
    dtX2 = G - X*GX';
end 
Q = 1; Cval = F;  tau = opts.tau;

%% Print iteration header if debug == 1
if (opts.record == 1)
    fid = 1;
    fprintf(fid, '----------- Gradient_Retrac Method with Line search ----------- \n');
    fprintf(fid, '%4s %8s %8s %10s %10s\n', 'Iter', 'tau', 'F(X)', 'nrmG', 'XDiff');
    %fprintf(fid, '%4d \t %3.2e \t %3.2e \t %5d \t %5d	\t %6d	\n', 0, 0, F, 0, 0, 0);
end

%% main iteration
F_eval = zeros(opts.mxitr+1,1 );
Grad = zeros(opts.mxitr+1,1 );
F_eval(1) = F;
Grad(1) = nrmG; 


tstart = tic;
for itr = 1 : opts.mxitr
    XP = X;     FP = F;   dtXP = dtX;       
    
    % Descent Direction:
    if OrthG || beta == 0
        H = alpha*dtXP;
    else
        H = alpha*dtXP + beta*dtX2;
    end
    HtH = H'*H;  XHtH = 0.5*X*HtH;

    % Scale Step Size
    nls = 1; deriv = rho*nrmG^2;  
    while 1
        
        % Update:
        Y = XP - tau*H;         
        X = Y - (tau^2)*XHtH;
        if( norm( X'*X - eye(k),'fro' ) > 1e-13 )
             error_flag = 0;
             %eval('[U,~,V] = fastsvd(Y);', 'error_flag = 1;');
             eval('[U,~,V] = svd(Y,0);', 'error_flag = 1;');

            if(error_flag)
                [Qfactor,~] = qr(Y);
                X = Qfactor(:,1:k);
            else
                X = U(:,1:k)*V';
            end
        end
        
        % Calculate G, F,
        [F, G] = feval(fun, X, varargin{:});
        out.nfe = out.nfe + 1;
        
        if F <= Cval - tau*deriv || nls >= 5
            break;
        end
        tau = eta*tau; nls = nls + 1;
        
    end  
    
    GX = G'*X;   dtX = G - X*GX;
    nrmG  = norm(dtX, 'fro');
    if ~OrthG && beta ~= 0 
        dtX2 = G - X*GX';   
    end
    F_eval(itr+1) = F; Grad(itr+1) = nrmG;

    % BB-Step
    S = X - XP;  XDiff = norm(S,'fro')/sqrt(n);
    tau = opts.tau;  FDiff = abs(FP-F)/(abs(FP)+1);
    
    if iscomplex
        Y = dtX - dtXP; SY = abs(sum(sum(conj(S).*Y)));
        if mod(itr,2)==0; tau = sum(sum(conj(S).*S))/SY; 
        else tau = SY/sum(sum(conj(Y).*Y)); end    
    else
        Y = dtX - dtXP;     SY = abs(sum(sum(S.*Y)));
        if mod(itr,2)==0; tau = sum(sum(S.*S))/SY;
        else tau  = SY/sum(sum(Y.*Y)); end        
    end
    
    tau = max(min(tau, 1e20), 1e-20);
    
    if (record >= 1)
        fprintf('%4d  %3.2e  %4.3e  %3.2e  %3.2e  %3.2e  %2d\n', ...
            itr, tau, F, nrmG, XDiff, FDiff, nls);
    end
    
    % Termination Rules.
    crit(itr,:) = [nrmG, XDiff, FDiff];
    mcrit = mean(crit(itr-min(nt,itr)+1:itr, :),1);   
    if ( XDiff < xtol && FDiff < ftol ) || nrmG < gtol || all(mcrit(2:3) < 10*[xtol, ftol])  
        if itr <= 2
            ftol = 0.1*ftol;
            xtol = 0.1*xtol;
            gtol = 0.1*gtol;
        else
            out.msg = 'converge';
            break;
        end
    end
    
    % Update: Parameters Non-monotone search
    Qp = Q; Q = gamma*Qp + 1; Cval = (gamma*Qp*Cval + F)/Q;

 end
time = toc(tstart);

F_eval = F_eval(1:itr,1);
Grad = Grad(1:itr,1);

if itr >= opts.mxitr
    out.msg = 'exceed max iteration';
end

out.feasi = norm(X'*X - eye(k),'fro');
if  out.feasi > 1e-13
    X = MGramSchmidt(X);
    [F, ~] = feval(fun, X, varargin{:});
    out.nfe = out.nfe + 1;
    out.feasi = norm(X'*X - eye(k),'fro');
end

out.nrmG = nrmG;
out.fval = F;
out.itr = itr;
out.time = time;
end







